module CrossSpec
  class SpecWaiting
    def initialize(tasks)
      @positive_matches = 0
      @data = []
      @tasks = Array(tasks)
    end

    def process(message, &block)
      return unless message.is_a?(TaskMessage)
      return unless @tasks.include?(message.task)

      positive_match = if block
        yield message.data, message.biomarkers
      elsif DistributedTracing.get.any?
        (message.biomarkers & DistributedTracing.get).any?
      else
        false
      end

      if positive_match
        increment_matches
        store_data(message)
      end
    end

    def done?
      @positive_matches === @tasks.length
    end

    def data
      @data
    end

    private

    def increment_matches
      @positive_matches += 1
    end

    def store_data(message)
      i = @tasks.index(message.task)
      @data[i] = message.data
    end
  end

  class Spec
    def self.trace(&block)
      spec = new
      DistributedTracing.trace do
        spec.instance_eval(&block)
      end
    end

    def async(service:, data: nil)
      DistributedTracing.trace do
        str = JSON.dump(
          service: service,
          biomarkers: DistributedTracing.get,
          data: data
        )
        CrossSpec.broadcast(str)
      end
      nil
    end

    def sync(service:, data: nil, tasks:, options: {}, &block)
      async(service: service, data: data)
      await(tasks: tasks, options: options, &block)
    end

    def await(tasks:, options: {}, &block)
      combined = { timeout: 5 }.merge(options)
      waiting = SpecWaiting.new(tasks)
      ::Timeout::timeout(combined[:timeout]) do
        CrossSpec.listen! do |message|
          waiting.process(message, &block)
          return waiting.data.length == 1 ? waiting.data.first : waiting.data if waiting.done?
          nil # nil keeps blocking listen!
        end
      end
    end
  end
end
